import collections
import functools
import typing
from collections.abc import Sequence

import jax.numpy as jnp
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from qfi_opt import spin_models

if typing.TYPE_CHECKING:
    from mpl_toolkits import mplot3d

try:
    import cmocean

    def sphere_cmap(color_vals: Sequence[float]) -> np.ndarray:
        return cmocean.cm.amp(color_vals)

except ModuleNotFoundError:

    def sphere_cmap(color_vals: Sequence[float]) -> np.ndarray:
        return plt.get_cmap("inferno")(color_vals)


def get_spin_length_projections(state: np.ndarray | jnp.ndarray) -> dict[float, np.ndarray]:
    """Compute the projections of a state onto manifolds of fixed spin length S.

    More specifically, compute the dictionary `{S: state_S for S in spin_length_vals}`, where
    - `state_S` = `P_S state P_S`, and
    - `P_S` is a projector onto the manifold with fixed spin length S.
    """
    num_qubits = spin_models.log2_int(state.shape[0])
    spin_length_projectors = get_spin_length_projectors(num_qubits)
    return {val: proj @ state @ proj for val, proj in spin_length_projectors.items()}


@functools.cache
def get_spin_length_projectors(num_qubits: int) -> dict[float, np.ndarray]:
    """Construct projectors onto manifolds of fixed spin length S."""
    spin_x, spin_y, spin_z = spin_models.collective_spin_ops(num_qubits)
    vals, vecs = np.linalg.eigh(spin_x @ spin_x + spin_y @ spin_y + spin_z @ spin_z)
    vals = np.round(np.sqrt(4 * vals + 1) - 1) / 2

    projectors: dict[float, np.ndarray] = collections.defaultdict(lambda: np.zeros((2**num_qubits,) * 2, dtype=complex))
    for spin_length_val, spin_length_vec in zip(vals, vecs.T):
        projectors[spin_length_val] += np.outer(spin_length_vec, spin_length_vec.conj())
    return projectors


def axis_spin_op(theta: float, phi: float, num_qubits: int) -> np.ndarray:
    """Construct the spin operator oriented along a given axis."""
    spin_x, spin_y, spin_z = spin_models.collective_spin_ops(num_qubits)
    return np.cos(theta) * spin_z + np.sin(theta) * (np.cos(phi) * spin_x + np.sin(phi) * spin_y)


def get_polarization(state_projections: dict[float, np.ndarray], theta: float, phi: float, cutoff: float = 1e-3) -> float:
    """Compute the polarization of a given state in the given direction.

    The polarization is defined by the average over all Husimi probability distribution functions (averaged over manifolds with fixed spin length S).

    The input state is a dictionary of the full state's projections onto manifolds of fixed spin length S.
    """
    num_qubits = spin_models.log2_int(next(iter(state_projections.values())).shape[0])
    spin_op = axis_spin_op(theta, phi, num_qubits)
    spin_op_vals, spin_op_vecs = np.linalg.eigh(spin_op)

    polarization = 0
    for spin_length_val, state_projection in state_projections.items():
        weight = 2 * spin_length_val + 1  # normalization factor for Husimi probability distribution
        if weight * np.trace(state_projection) < cutoff:
            continue

        max_spin_indices = np.isclose(spin_op_vals, spin_length_val)
        for spin_op_vec in spin_op_vecs[:, max_spin_indices].T:
            polarization += weight * (spin_op_vec.conj() @ state_projection @ spin_op_vec).real

    return polarization


def husimi(
    state: np.ndarray | jnp.ndarray,
    grid_size: int = 101,
    single_sphere: bool = True,
    figsize: tuple[float, float] | None = None,
    rasterized: bool = True,
    view_angles: tuple[float, float] = (0, 0),
    shade: bool = False,
    color_max: float | None = None,
) -> tuple[mpl.figure.Figure, list[mplot3d.axes3d.Axes3D]]:
    if figsize is None:
        figsize = plt.figaspect(1 if single_sphere else 0.5)

    # initialize grid and color map

    theta, phi = np.meshgrid(np.linspace(0, np.pi, grid_size), np.linspace(0, 2 * np.pi, grid_size))
    z_vals = np.cos(theta)
    x_vals = np.sin(theta) * np.cos(phi)
    y_vals = np.sin(theta) * np.sin(phi)

    state_projections = get_spin_length_projections(state)
    color_vals = np.vectorize(functools.partial(get_polarization, state_projections))(theta, phi)
    vmax = np.max(abs(color_vals)) if not color_max else color_max
    norm = mpl.colors.Normalize(vmax=vmax, vmin=0)
    color_map = sphere_cmap(norm(color_vals))

    # plot sphere

    figure = plt.figure(figsize=figsize)
    axes: list[mplot3d.axes3d.Axes3D]
    if single_sphere:
        axes = [figure.add_subplot(111, projection="3d")]
    else:
        axes = [figure.add_subplot(121, projection="3d"), figure.add_subplot(122, projection="3d")]

    for axis, side in zip(axes, [+1, -1]):
        axis.plot_surface(side * x_vals, side * y_vals, z_vals, rstride=1, cstride=1, facecolors=color_map, rasterized=rasterized, shade=shade)

    # clean up figure

    elev, azim = view_angles
    for axis in axes:
        axis.set_xlim((-0.7, 0.7))
        axis.set_ylim((-0.7, 0.7))
        axis.set_zlim((-0.8, 0.8))
        axis.view_init(elev=elev, azim=azim)
        axis.set_axis_off()

    left = -0.01
    right = 1
    bottom = -0.03
    top = 1
    rect = (left, bottom, right, top)
    figure.tight_layout(pad=0, w_pad=0, h_pad=0, rect=rect)
    return figure, axes


def histogram(
    state: np.ndarray | jnp.ndarray,
    figsize: tuple[int, int] = (4, 3),
) -> tuple[mpl.figure.Figure, mpl.axes._axes.Axes]:
    """Plot the distribution function over collective-Sz measurement outcomes."""
    num_spins = spin_models.log2_int(state.shape[0])
    state_tensor = np.reshape(state, (2,) * num_spins * 2)

    # collect the probabilities for different numbers of |1> states

    probabilities = np.zeros(num_spins + 1)
    for bitstring in np.ndindex((2,) * num_spins):
        num_spins_up = num_spins - sum(bitstring)  # number of zeros in the bitstring
        tensor_index = bitstring + bitstring
        probabilities[num_spins_up] += state_tensor[tensor_index].real

    # plot a histogram and return

    measurement_outcomes = np.arange(num_spins + 1) - num_spins / 2

    figure, axis = plt.subplots(figsize=figsize)
    axis.bar(measurement_outcomes, probabilities, width=0.9)
    axis.set_ylabel("Probability")
    axis.set_xlabel("Axial spin projection")
    figure.tight_layout()
    return figure, axis



#!/usr/bin/env python3
import matplotlib.pyplot as plt
import numpy as np

from qfi_opt import spin_models
from qfi_opt.examples.calculate_qfi import compute_eigendecompotion, compute_QFI

num_spins = 4
dissipation = 0
op = spin_models.collective_op(spin_models.PAULI_Z, num_spins) / (2 * num_spins)

num_pts = 11
x_ = np.linspace(0.0, 1.0, num_pts)
# x, y, z = np.meshgrid(x_, x_, x_, indexing="ij")
y, z = np.meshgrid(x_, x_, indexing="ij")
obj_vals = np.zeros_like(y)

for i in range(num_pts):
    print(i, flush=True)
    for j in range(num_pts):
        params = np.array([0.5, y[i, j], z[i, j], 0])
        rho = spin_models.simulate_OAT(params, num_spins, dissipation_rates=dissipation)
        vals, vecs = compute_eigendecompotion(rho)
        qfi = compute_QFI(vals, vecs, op)
        obj_vals[i, j] = qfi

fig, ax = plt.subplots()

CS = ax.contour(y, z, obj_vals)
cbar = plt.colorbar(CS)
cbar.set_label("QFI")

plt.title(f"max QFI = {np.max(obj_vals)}")

plt.savefig("contours_with_first_param_half_last_param_zero_" + str(num_pts) + ".png", dpi=300)



#!/usr/bin/env python3
import os
import sys

import nlopt
import numpy as np
from ibcdfo.pounders import pounders
from ibcdfo.pounders.general_h_funs import identity_combine as combinemodels

import qfi_opt
from qfi_opt import spin_models
from qfi_opt.examples.calculate_qfi import compute_eigendecomposition, compute_QFI

sys.path.append("../../")

root_dir = os.path.dirname(os.path.dirname(qfi_opt.__file__))
minq5_dir = os.path.join(root_dir, "minq", "py", "minq5")
if not os.path.isdir(minq5_dir):
    messages = [
        "Please install (or symlink) MINQ in the QFI-Opt project directory.",
        "You can do this with:",
        f"\n  git clone https://github.com/POptUS/MINQ.git {root_dir}/minq\n",
        "Or, if you already have MINQ somewhere, run:",
        f"\n  ln -s <minq-path> {root_dir}/minq\n",
    ]
    exit("\n".join(messages))
sys.path.append(minq5_dir)


def sim_wrapper(x, h, qfi_grad, obj, obj_params):
    use_DB = False
    match = 0
    if use_DB:
        # Look through the database to see if there is a match
        database = obj.__name__ + "_" + str(obj_params["N"]) + "_" + str(obj_params["dissipation"]) + "_database.npy"
        DB = []
        if os.path.exists(database):
            DB = np.load(database, allow_pickle=True)
            for db_entry in DB:
                if np.allclose(db_entry["var_vals"], x, rtol=1e-12, atol=1e-12):
                    rho = db_entry["rho"]
                    match = 1
                    break

    if match == 0:
        # Do the sim
        rho = obj(x, obj_params["N"], dissipation_rates=obj_params["dissipation"])

        if use_DB:
            # Update database
            to_save = {"rho": rho, "var_vals": x}
            DB = np.append(DB, to_save)
            np.save(database, DB)

    num_parameters = len(x)
    dA = np.zeros((num_parameters, N**2, N**2), dtype="cdouble")
    d2A = np.zeros((num_parameters, N**2, N**2), dtype="cdouble")
    if qfi_grad.size > 0:
        # Approximate the gradient
        I = np.eye(num_parameters)

        for parameter in range(num_parameters):
            # compute dA with respect to parameter
            rho_p = obj(x + h * I[parameter], obj_params["N"], dissipation_rates=obj_params["dissipation"])
            rho_m = obj(x - h * I[parameter], obj_params["N"], dissipation_rates=obj_params["dissipation"])
            dA[parameter] = (rho_p - rho_m) / (2 * h)
            # compute d2A with respect to parameter
            d2A[parameter] = (rho_p - 2 * rho + rho_m) / (h**2)

    # Compute eigendecomposition
    vals, vecs = compute_eigendecomposition(rho)

    qfi, new_grad = compute_QFI(vals, vecs, obj_params["G"], rho, dA, d2A, qfi_grad)
    print(x, qfi, new_grad, flush=True)

    try:
        if qfi_grad.size > 0:
            qfi_grad[:] = -1.0 * new_grad
    except:
        qfi_grad[:] = []

    return -1.0 * qfi  # , -qfi_grad  # negative because we are maximizing
    # else:
    #    return -1 * qfi


def quick_test_gradient(obj, obj_params, x):
    h = 1e-8
    calfun = lambda x: sim_wrapper(x, h, True, obj, obj_params)
    calfun_f = lambda x: sim_wrapper(x, h, False, obj, obj_params)
    f0, g = calfun(x)
    print("f0: ", f0)

    # check for descent in gradient direction
    for j in range(4, 18, 2):
        stepsize = 10 ** (-j)
        xp = x - stepsize * g
        fp = calfun_f(xp)
        print("fp: ", fp, "stepsize: ", stepsize)


def run_pounder(obj, obj_params, n, x0):
    h = 1e-6
    calfun = lambda x: sim_wrapper(x, h, False, obj, obj_params)
    X = np.array(x0)
    F = np.array(calfun(X))
    Low = -np.inf * np.ones((1, n))
    Upp = np.inf * np.ones((1, n))
    mpmax = 2 * n + 1
    delta = 0.1
    m = 1
    nfs = 1
    printf = False
    spsolver = 2
    gtol = 1e-9
    xind = 0
    hfun = lambda F: F

    X, F, _, xkin = pounders(calfun, X, n, mpmax, max_evals, gtol, delta, nfs, m, F, xind, Low, Upp, printf, spsolver, hfun, combinemodels)

    # print("optimum at ", X[xkin])
    # print("minimum value = ", F[xkin])

    return F[xkin], X[xkin]


def run_nlopt(obj, obj_params, num_params, x0, solver):
    opt = nlopt.opt(getattr(nlopt, solver), num_params)
    # opt = nlopt.opt(nlopt.LN_NELDERMEAD, num_params)  # Doesn't use derivatives and will work
    # opt = nlopt.opt(nlopt.LD_MMA, num_params) # Needs derivatives to work. Without grad being set (in-place) it is zero, so first iterate is deemed stationary

    h = 1e-5
    opt.set_min_objective(lambda x, grad: sim_wrapper(x, h, grad, obj, obj_params))
    opt.set_xtol_rel(1e-4)
    opt.set_maxeval(300)
    opt.set_lower_bounds(-10.0 * np.ones(num_params))
    opt.set_upper_bounds(10.0 * np.ones(num_params))
    opt.set_vector_storage(0)

    # # Because the objective is periodic, don't set bounds (but don't need to sample so much)
    # opt.set_lower_bounds(lb)
    # opt.set_upper_bounds(ub)

    x = opt.optimize(x0)
    minf = opt.last_optimum_value()
    # print("optimum at ", x)
    # print("minimum value = ", minf)
    # print("result code = ", opt.last_optimize_result())

    return minf, x


if __name__ == "__main__":  # noqa: C901 # ignore "complexity" check for the code below
    # comm = MPI.COMM_WORLD
    # rank = comm.Get_rank()
    # size = comm.Get_size()
    size = 1

    N = 4
    G = spin_models.collective_op(spin_models.PAULI_Z, N) / (2 * N)

    for dissipation_rate in np.append([0], np.linspace(0.1, 5, 20)):
        obj_params = {}
        obj_params["N"] = N
        obj_params["dissipation"] = dissipation_rate
        obj_params["G"] = G

        max_evals = 300

        seed = 88
        np.random.seed(seed)

        for num_params in [4, 5]:
            lb = np.zeros(num_params)
            ub = np.ones(num_params)

            # x0 = 0.5 * np.ones(num_params)
            x0 = np.random.uniform(lb, ub, num_params)

            match num_params:
                case 4:
                    models = ["simulate_OAT", "simulate_ising_chain", "simulate_XX_chain"]
                    # models = ["simulate_OAT"]
                case 5:
                    models = ["simulate_TAT", "simulate_local_TAT_chain"]
                    # models = ["simulate_TAT"]

            for model in models:
                obj = getattr(spin_models, model)
                minf, xfinal = run_nlopt(obj, obj_params, num_params, x0, "LD_LBFGS")
                # h = 1e-6
                # calfun = lambda x, grad: sim_wrapper(x, h, grad, obj, obj_params)
                # minf, xfinal = run_nlopt(obj, obj_params, num_params, x0, "LN_BOBYQA")
                # grad = np.zeros(num_params)
                # calfun(xfinal, grad)



from __future__ import annotations

from collections.abc import Callable
from typing import Literal

import jax.numpy as jnp
import numpy as np

ARRAY_TYPE = np.ndarray | jnp.ndarray


class Dissipator:
    """Data structure to represent a dissipation operator.

    Only (uncorrelated) single-qubit dissipation is supported at the moment.

    The Dissipator class overloads matrix multiplication, such that the time derivative of a density
    matrix 'rho' evolving only under dissipation is 'dissipator @ rho'.

    A Dissipator is initialized with the following arguments:
    (1) Three dissipation rates.  If only a single dissipation rate is provided, then all
        dissipation rates are assumed to be equal.
    (2) A string that specifies a dissipation "format".  The dissipation format may be one of the
        following:
        - "XYZ": The dissipation rates respectively correspond to dephasing along the X, Y, and Z
            axes, with jump operators X/2, Y/2, and Z/2.  Here X, Y, and Z are Pauli operators.  If
            all dissipation rates are equal, then XYZ-type dissipation generates the depolarizing
            channel.
        - "PMZ": The dissipation rates respectively correspond to spontaneous spin excitation,
            decay, and dephasing, with jump operators P = (X+iY)/2, M = (X-iY)/2, and Z/2.
            PMZ-type dissipation generates the depolarizing channel when dissipation rates are
            proportional to (1/2, 1/2, 1).
    """

    def __init__(self, dissipation_rates: float | tuple[float, float, float], dissipation_format: str = "XYZ") -> None:
        if not isinstance(dissipation_rates, tuple):
            dissipation_rates = (dissipation_rates,) * 3
        assert all(rate >= 0 for rate in dissipation_rates), "dissipation rates cannot be negative!"
        self._bare_rates = dissipation_rates
        self._format = dissipation_format

        # This dissipator acts on a density matrix as
        #   dissipator @ rho = sum_j rate_j sum(qubit_term_j(rho, qubit) for qubit in all_qubits).
        # Identify the rates and qubit_term functions to use below.

        rates = [0.0, 0.0, 0.0]
        if dissipation_format == "XYZ":
            terms = (term_XYZ_1, term_XYZ_2, term_XYZ_3)
            rate_sx, rate_sy, rate_sz = dissipation_rates
            rates[0] = (rate_sx + rate_sy) / 4 + rate_sz / 2
            rates[1] = (rate_sx + rate_sy) / 4
            rates[2] = (rate_sx - rate_sy) / 4

        elif dissipation_format == "PMZ":
            terms = (term_PMZ_1, term_PMZ_2, term_PMZ_3)
            rate_sp, rate_sm, rate_sz = dissipation_rates
            rates[0] = sum(dissipation_rates) / 2
            rates[1] = rate_sp
            rates[2] = rate_sm

        else:
            raise ValueError(f"dissipation format not recognized: {dissipation_format}")

        self._rates = []
        self._terms = []
        for rate, term in zip(rates, terms):
            if rate:
                self._rates.append(rate)
                self._terms.append(term)

    def __matmul__(self, density_op: ARRAY_TYPE) -> ARRAY_TYPE | Literal[0]:
        return sum(rate * term(density_op) for rate, term in zip(self._rates, self._terms))  # type: ignore[return-value,misc]

    def __mul__(self, scalar: float) -> Dissipator:
        bare_rates = (
            scalar * self._bare_rates[0],
            scalar * self._bare_rates[1],
            scalar * self._bare_rates[2],
        )
        return Dissipator(bare_rates, self._format)

    def __rmul__(self, scalar: float) -> Dissipator:
        return self * scalar

    def __truediv__(self, scalar: float) -> Dissipator:
        return self * (1 / scalar)

    def __bool__(self) -> bool:
        return bool(sum(self._rates))


def log2_int(val: int) -> int:
    return val.bit_length() - 1


"""
For any specified qubit 'q', the density matrix 'rho' can be written in the form
  rho = ⌈ rho_q_00, rho_q_01 ⌉
        ⌊ rho_q_10, rho_q_11 ⌋,
where 'rho_q_ab = <a|rho|b>_q = Tr(rho |b><a|_q)' is a block of 'rho'.

The methods below accept a density matrix reshaped in such a way that
  rho[:, a, :, b, :] = rho_q_ab.
These methods then manipulate blocks of row to perform various single-qubit transformations.

To reduce boilerplate comments, these methods below will only specify how the blocks get rearranged.
"""


def qubit_term_XYZ_1(density_op: ARRAY_TYPE) -> ARRAY_TYPE:
    """Starting with the matrix ⌈ a, b ⌉  return the matrix ⌈  0, -b ⌉
    ...                         ⌊ c, d ⌋,                   ⌊ -c,  0 ⌋.
    """
    output: ARRAY_TYPE
    if isinstance(density_op, jnp.ndarray):
        output = jnp.empty_like(density_op)
        output = output.at[:, 0, :, 0, :].set(0)
        output = output.at[:, 0, :, 1, :].set(-density_op[:, 0, :, 1, :])
        output = output.at[:, 1, :, 0, :].set(-density_op[:, 1, :, 0, :])
        output = output.at[:, 1, :, 1, :].set(0)
    else:
        output = np.empty_like(density_op)
        output[:, 0, :, 0, :] = 0
        output[:, 0, :, 1, :] = -density_op[:, 0, :, 1, :]
        output[:, 1, :, 0, :] = -density_op[:, 1, :, 0, :]
        output[:, 1, :, 1, :] = 0
    return output


def qubit_term_XYZ_2(density_op: ARRAY_TYPE) -> ARRAY_TYPE:
    """Starting with the matrix ⌈ a, b ⌉  return the matrix ⌈ d-a,  0  ⌉
    ...                         ⌊ c, d ⌋,                   ⌊  0,  a-d ⌋.
    """
    output: ARRAY_TYPE
    if isinstance(density_op, jnp.ndarray):
        output = jnp.empty_like(density_op)
        output = output.at[:, 0, :, 0, :].set(density_op[:, 1, :, 1, :] - density_op[:, 0, :, 0, :])
        output = output.at[:, 0, :, 1, :].set(0)
        output = output.at[:, 1, :, 0, :].set(0)
        output = output.at[:, 1, :, 1, :].set(-output[:, 0, :, 0, :])
    else:
        output = np.empty_like(density_op)
        output[:, 0, :, 0, :] = density_op[:, 1, :, 1, :] - density_op[:, 0, :, 0, :]
        output[:, 0, :, 1, :] = 0
        output[:, 1, :, 0, :] = 0
        output[:, 1, :, 1, :] = -output[:, 0, :, 0, :]
    return output


def qubit_term_XYZ_3(density_op: ARRAY_TYPE) -> ARRAY_TYPE:
    """Starting with the matrix ⌈ a, b ⌉  return the matrix ⌈ 0, c ⌉
    ...                         ⌊ c, d ⌋,                   ⌊ b, 0 ⌋.
    """
    output: ARRAY_TYPE
    if isinstance(density_op, jnp.ndarray):
        output = jnp.empty_like(density_op)
        output = output.at[:, 0, :, 0, :].set(0)
        output = output.at[:, 0, :, 1, :].set(density_op[:, 1, :, 0, :])
        output = output.at[:, 1, :, 0, :].set(density_op[:, 0, :, 1, :])
        output = output.at[:, 1, :, 1, :].set(0)
    else:
        output = np.empty_like(density_op)
        output[:, 0, :, 0, :] = 0
        output[:, 0, :, 1, :] = density_op[:, 1, :, 0, :]
        output[:, 1, :, 0, :] = density_op[:, 0, :, 1, :]
        output[:, 1, :, 1, :] = 0
    return output


qubit_term_PMZ_1 = qubit_term_XYZ_1


def qubit_term_PMZ_2(density_op: ARRAY_TYPE) -> ARRAY_TYPE:
    """Starting with the matrix ⌈ a, b ⌉  return the matrix ⌈ d,  0 ⌉
    ...                         ⌊ c, d ⌋,                   ⌊ 0, -d ⌋.
    """
    output: ARRAY_TYPE
    if isinstance(density_op, jnp.ndarray):
        output = jnp.empty_like(density_op)
        output = output.at[:, 0, :, 0, :].set(density_op[:, 1, :, 1, :])
        output = output.at[:, 0, :, 1, :].set(0)
        output = output.at[:, 1, :, 0, :].set(0)
        output = output.at[:, 1, :, 1, :].set(-density_op[:, 1, :, 1, :])
    else:
        output = np.empty_like(density_op)
        output[:, 0, :, 0, :] = density_op[:, 1, :, 1, :]
        output[:, 0, :, 1, :] = 0
        output[:, 1, :, 0, :] = 0
        output[:, 1, :, 1, :] = -density_op[:, 1, :, 1, :]
    return output


def qubit_term_PMZ_3(density_op: ARRAY_TYPE) -> ARRAY_TYPE:
    """Starting with the matrix ⌈ a, b ⌉  return the matrix ⌈ -a,  0 ⌉
    ...                         ⌊ c, d ⌋,                   ⌊  0,  a ⌋.
    """
    output: ARRAY_TYPE
    if isinstance(density_op, jnp.ndarray):
        output = jnp.empty_like(density_op)
        output = output.at[:, 0, :, 0, :].set(-density_op[:, 0, :, 0, :])
        output = output.at[:, 0, :, 1, :].set(0)
        output = output.at[:, 1, :, 0, :].set(0)
        output = output.at[:, 1, :, 1, :].set(density_op[:, 0, :, 0, :])
    else:
        output = np.empty_like(density_op)
        output[:, 0, :, 0, :] = -density_op[:, 0, :, 0, :]
        output[:, 0, :, 1, :] = 0
        output[:, 1, :, 0, :] = 0
        output[:, 1, :, 1, :] = density_op[:, 0, :, 0, :]
    return output


def sum_qubit_terms(
    qubit_term: Callable[[ARRAY_TYPE], ARRAY_TYPE],
) -> Callable[[ARRAY_TYPE], ARRAY_TYPE | Literal[0]]:
    """For a given single-qubit transformation, sum over its action on all qubits."""

    def sum_of_qubit_terms(density_op: ARRAY_TYPE) -> ARRAY_TYPE | Literal[0]:
        num_qubits = log2_int(density_op.size) // 2
        input_shape = density_op.shape
        output: ARRAY_TYPE | Literal[0] = 0
        for qubit in range(num_qubits):
            dim_a = 2**qubit
            dim_b = 2 ** (num_qubits - qubit - 1)
            density_op = density_op.reshape((dim_a, 2, dim_b * dim_a, 2, dim_b))
            output += qubit_term(density_op).reshape(input_shape)
        density_op = density_op.reshape(input_shape)
        return output

    return sum_of_qubit_terms


term_XYZ_1 = sum_qubit_terms(qubit_term_XYZ_1)
term_XYZ_2 = sum_qubit_terms(qubit_term_XYZ_2)
term_XYZ_3 = sum_qubit_terms(qubit_term_XYZ_3)
term_PMZ_1 = sum_qubit_terms(qubit_term_PMZ_1)
term_PMZ_2 = sum_qubit_terms(qubit_term_PMZ_2)
term_PMZ_3 = sum_qubit_terms(qubit_term_PMZ_3)



#!/usr/bin/env python3
import argparse
import functools
import itertools
import os
import sys
from collections.abc import Callable, Sequence

import numpy

from qfi_opt.dissipation import Dissipator

DISABLE_DIFFRAX = bool(os.getenv("DISABLE_DIFFRAX"))

if not DISABLE_DIFFRAX:
    import diffrax
    import jax
    import jax.numpy as np

    jax.config.update("jax_enable_x64", True)

else:
    import numpy as np  # type: ignore[no-redef]
    import scipy

COMPLEX_TYPE = np.complex128
DEFAULT_DISSIPATION_FORMAT = "XYZ"

# qubit/spin states
KET_0 = np.array([1, 0], dtype=COMPLEX_TYPE)  # |0>, spin up
KET_1 = np.array([0, 1], dtype=COMPLEX_TYPE)  # |1>, spin down

# Pauli operators
PAULI_I = np.array([[1, 0], [0, 1]], dtype=COMPLEX_TYPE)  # |0><0| + |1><1|
PAULI_Z = np.array([[1, 0], [0, -1]], dtype=COMPLEX_TYPE)  # |0><0| - |1><1|
PAULI_X = np.array([[0, 1], [1, 0]], dtype=COMPLEX_TYPE)  # |0><1| + |1><0|
PAULI_Y = -1j * PAULI_Z @ PAULI_X


def log2_int(val: int) -> int:
    return val.bit_length() - 1


def simulate_sensing_protocol(
    params: Sequence[float] | np.ndarray,
    entangling_hamiltonian: np.ndarray,
    *,
    dissipation_rates: float | tuple[float, float, float] = 0.0,
    dissipation_format: str = DEFAULT_DISSIPATION_FORMAT,
    axial_symmetry: bool = False,
) -> np.ndarray:
    """Simulate a sensing protocol, and return the final state (density matrix).

    Starting with an initial all-|1> state (all spins pointing down along the Z axis):
    1. Rotate about an axis in the XY plane.
    2. Evolve under a given entangling Hamiltonian.
    3. Rotate about the X axis.
    4. Rotate about the Y axis.

    Step 1 rotates by an angle 'np.pi * params[0]', about the axis 'np.pi * params[1]'.
    Step 2 evolves under the given entangling Hamiltonian for time 'params[2] * np.pi * num_qubits'.
    Step 3 rotates by an angle 'np.pi * params[3]'.
    Step 3 rotates by an angle 'np.pi * params[-1]'.

    If axial_symmetry is set to True, 0 is appended to the parameter list.
    Afterwards, if there are more than 5 parameters, steps 2 and 3 are repeated as appropriate.

    If dissipation_rates is nonzero, qubits experience dissipation during the entangling steps.
    See the documentation for the Dissipator class for a general explanation of the
    dissipation_rates and dissipation_format arguments.

    This method additionally divides the dissipator (equivalently, all dissipation rates) by a
    factor of 'np.pi * num_qubits' in order to "normalize" dissipation time scales, and make them
    comparable to the time scales of coherent evolution.  Dividing a Dissipator with homogeneous
    dissipation rates 'r' by a factor of 'np.pi * num_qubits' makes so that each qubit depolarizes
    with probability 'e^(-params[0] * r)' by the end of the OAT protocol.
    """
    if len(params) < 5 or not len(params) % 2:
        raise ValueError(f"The number of parameters should be an odd number >=5, not {len(params)}.")

    num_qubits = log2_int(entangling_hamiltonian.shape[0])

    # rotate the all-|1> state about a chosen axis
    time = params[0] * np.pi
    cos = np.cos(time / 2)
    sin = np.sin(time / 2)
    axis_angle = params[1] * np.pi
    qubit_ket = cos * KET_1 - 1j * np.exp(-1j * axis_angle) * sin * KET_0
    qubit_state = np.outer(qubit_ket, qubit_ket.conj())
    state = functools.reduce(np.kron, [qubit_state] * num_qubits)

    # entangle - rotate layers
    dissipator = Dissipator(dissipation_rates, dissipation_format) / (np.pi * num_qubits)
    for pp in range(2, len(params) - 1, 2):
        # entangle
        time = params[pp] * np.pi * num_qubits
        state = evolve_state(state, time, entangling_hamiltonian, dissipator)

        # rotate about Sx
        time = params[pp + 1] * np.pi
        mat_rx = np.cos(time / 2) * PAULI_I - 1j * np.sin(time / 2) * PAULI_X
        state = apply_globally(state, mat_rx, num_qubits)

    # rotate about Sy
    time = params[-1] * np.pi
    mat_ry = np.cos(time / 2) * PAULI_I - 1j * np.sin(time / 2) * PAULI_Y
    state = apply_globally(state, mat_ry, num_qubits)

    return state


def apply_globally(density_op: np.ndarray, qubit_op: np.ndarray, num_qubits: int) -> np.ndarray:
    """Apply the given qubit operator to all qubits of a density operator."""
    qubit_op_dag = qubit_op.conj().T
    for qubit in range(num_qubits):
        dim_a = 2**qubit
        dim_b = 2 ** (num_qubits - qubit - 1)
        density_op = density_op.reshape((dim_a, 2, dim_b * dim_a, 2, dim_b))
        density_op = np.einsum("ij,AjBkC,kl->AiBlC", qubit_op, density_op, qubit_op_dag)
    return density_op.reshape((2**num_qubits,) * 2)


def enable_axial_symmetry(simulate_func: Callable[..., np.ndarray]) -> Callable[..., np.ndarray]:
    """Decorator to enable an axially-symmetric version of a simulation method.

    Axial symmetry means that the last parameter can be set to zero without loss of generality.
    """

    def simulate_func_with_symmetry(params: Sequence[float] | np.ndarray, *args: object, **kwargs: object) -> np.ndarray:
        if len(params) == 4:
            # Verify that the dissipation arguments are compatible with axial symmetry.
            dissipation_rates = kwargs.get("dissipation_rates", 0.0)
            dissipation_format = kwargs.get("dissipation_format", DEFAULT_DISSIPATION_FORMAT)
            if dissipation_format == "XYZ" and hasattr(dissipation_rates, "__iter__"):
                rate_sx, rate_sy, *_ = dissipation_rates
                if not rate_sx == rate_sy:
                    raise ValueError(
                        f"Dissipation format {dissipation_format} with rates {dissipation_rates} does not respect axial symmetry."
                        "\nPlease provide at least 5 parameters, or pick different options."
                    )
            params = np.append(np.array(params), 0.0)

        return simulate_func(params, *args, **kwargs)

    return simulate_func_with_symmetry


@enable_axial_symmetry
def simulate_OAT(
    params: Sequence[float] | np.ndarray,
    num_qubits: int,
    *,
    dissipation_rates: float | tuple[float, float, float] = 0.0,
    dissipation_format: str = DEFAULT_DISSIPATION_FORMAT,
) -> np.ndarray:
    """Simulate a one-axis twisting (OAT) protocol."""
    _, _, collective_Sz = collective_spin_ops(num_qubits)
    hamiltonian = collective_Sz.diagonal() ** 2 / num_qubits
    return simulate_sensing_protocol(
        params,
        hamiltonian,
        dissipation_rates=dissipation_rates,
        dissipation_format=dissipation_format,
    )


def simulate_TAT(
    params: Sequence[float] | np.ndarray,
    num_qubits: int,
    *,
    dissipation_rates: float | tuple[float, float, float] = 0.0,
    dissipation_format: str = DEFAULT_DISSIPATION_FORMAT,
) -> np.ndarray:
    """Simulate a two-axis twisting (TAT) protocol."""
    collective_Sx, collective_Sy, _ = collective_spin_ops(num_qubits)
    hamiltonian = (collective_Sx @ collective_Sy + collective_Sy @ collective_Sx) / num_qubits
    return simulate_sensing_protocol(
        params,
        hamiltonian,
        dissipation_rates=dissipation_rates,
        dissipation_format=dissipation_format,
    )


def simulate_spin_chain(
    params: Sequence[float] | np.ndarray,
    num_qubits: int,
    coupling_op: np.ndarray,
    coupling_exponent: float = 0.0,
    *,
    dissipation_rates: float | tuple[float, float, float] = 0.0,
    dissipation_format: str = DEFAULT_DISSIPATION_FORMAT,
) -> np.ndarray:
    """Simulate an entangling protocol for a spin chain with power-law interactions."""
    normalization_factor = num_qubits * np.array([1 / abs(pp - qq) ** coupling_exponent for pp, qq in itertools.combinations(range(num_qubits), 2)]).mean()
    hamiltonian = sum(
        act_on_subsystem(num_qubits, coupling_op, pp, qq) / abs(pp - qq) ** coupling_exponent for pp, qq in itertools.combinations(range(num_qubits), 2)
    )
    return simulate_sensing_protocol(
        params,
        hamiltonian / normalization_factor,
        dissipation_rates=dissipation_rates,
        dissipation_format=dissipation_format,
    )


@enable_axial_symmetry
def simulate_ising_chain(
    params: Sequence[float] | np.ndarray,
    num_qubits: int,
    coupling_exponent: float = 0.0,
    *,
    dissipation_rates: float | tuple[float, float, float] = 0.0,
    dissipation_format: str = DEFAULT_DISSIPATION_FORMAT,
) -> np.ndarray:
    coupling_op = np.kron(PAULI_Z, PAULI_Z) / 2
    return simulate_spin_chain(
        params,
        num_qubits,
        coupling_op,
        coupling_exponent,
        dissipation_rates=dissipation_rates,
        dissipation_format=dissipation_format,
    )


@enable_axial_symmetry
def simulate_XX_chain(
    params: Sequence[float] | np.ndarray,
    num_qubits: int,
    coupling_exponent: float = 0.0,
    *,
    dissipation_rates: float | tuple[float, float, float] = 0.0,
    dissipation_format: str = DEFAULT_DISSIPATION_FORMAT,
) -> np.ndarray:
    coupling_op = (np.kron(PAULI_X, PAULI_X) + np.kron(PAULI_Y, PAULI_Y)) / 2
    return simulate_spin_chain(
        params,
        num_qubits,
        coupling_op,
        coupling_exponent,
        dissipation_rates=dissipation_rates,
        dissipation_format=dissipation_format,
    )


def simulate_local_TAT_chain(
    params: Sequence[float] | np.ndarray,
    num_qubits: int,
    coupling_exponent: float = 0.0,
    *,
    dissipation_rates: float | tuple[float, float, float] = 0.0,
    dissipation_format: str = DEFAULT_DISSIPATION_FORMAT,
) -> np.ndarray:
    coupling_op = (np.kron(PAULI_X, PAULI_Y) + np.kron(PAULI_Y, PAULI_X)) / 2
    return simulate_spin_chain(
        params,
        num_qubits,
        coupling_op,
        coupling_exponent,
        dissipation_rates=dissipation_rates,
        dissipation_format=dissipation_format,
    )


def evolve_state(
    density_op: np.ndarray,
    time: float | np.ndarray,
    hamiltonian: np.ndarray,
    dissipator: Dissipator | None = None,
    *,
    rtol: float = 1e-8,
    atol: float = 1e-8,
    disable_diffrax: bool = DISABLE_DIFFRAX,
    solver: diffrax.AbstractSolver | None = None,
    **diffrax_kwargs: object,
) -> np.ndarray:
    """Time-evolve a given initial density operator for a given amount of time under the given Hamiltonian and (optionally) Dissipator."""

    # treat negative times as evolving under the negative of the Hamiltonian
    # NOTE: this is required for autodiff to work
    if time.real < 0:
        time, hamiltonian = -time, -hamiltonian

    time_deriv = get_time_deriv(hamiltonian, dissipator)

    if not DISABLE_DIFFRAX:

        # set initial time step size, if necessary
        if "dt0" not in diffrax_kwargs:
            diffrax_kwargs["dt0"] = 0.002

        def _time_deriv(time: float, density_op: np.ndarray, hamiltonian: np.ndarray) -> np.ndarray:
            return time_deriv(time, density_op)

        term = diffrax.ODETerm(_time_deriv)
        solver = solver or diffrax.Tsit5()  # try also diffrax.Dopri8()
        solution = diffrax.diffeqsolve(term, solver, t0=0.0, t1=time, y0=density_op, args=(hamiltonian,), max_steps=None, **diffrax_kwargs)
        return solution.ys[-1]

    else:
        if np.isclose(time, 0, atol=atol):
            return density_op

        def scipy_time_deriv(time: float, density_op: np.ndarray) -> np.ndarray:
            density_op.shape = (hamiltonian.shape[0],) * 2  # type: ignore[misc]
            output = time_deriv(time, density_op)
            density_op.shape = (-1,)  # type: ignore[misc]
            return output.ravel()

        result = scipy.integrate.solve_ivp(
            scipy_time_deriv,
            [0, time],
            density_op.ravel(),
            t_eval=[time],
            rtol=rtol,
            atol=atol,
        )
        return result.y[:, -1].reshape(density_op.shape)


def get_time_deriv(
    hamiltonian: np.ndarray,
    dissipator: Dissipator | None = None,
    *,
    disable_diffrax: bool = DISABLE_DIFFRAX,
) -> Callable[[float, np.ndarray], np.ndarray]:
    """Construct a time derivative function that maps (state, time) --> d(state)/d(time)."""

    # construct the time derivative from coherent evolution
    if hamiltonian.ndim == 2:
        # ... with ordinary matrix multiplication
        def coherent_time_deriv(time: float, density_op: np.ndarray) -> np.ndarray:
            return -1j * (hamiltonian @ density_op - density_op @ hamiltonian)

    else:
        # 'hamiltonian' is a 1-D array of the values on the diagonal of the actual Hamiltonian,
        # so we can compute the commutator with array broadcasting, which is faster than matrix multiplication
        expanded_hamiltonian = np.expand_dims(hamiltonian, 1)

        def coherent_time_deriv(time: float, density_op: np.ndarray) -> np.ndarray:
            return -1j * (expanded_hamiltonian * density_op - density_op * hamiltonian)

    if not dissipator:
        return coherent_time_deriv

    def dissipative_time_deriv(time: float, density_op: np.ndarray) -> np.ndarray:
        return coherent_time_deriv(time, density_op) + dissipator @ density_op

    return dissipative_time_deriv


@functools.cache
def collective_spin_ops(num_qubits: int) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Construct collective spin operators."""
    return (
        collective_op(PAULI_X, num_qubits) / 2,
        collective_op(PAULI_Y, num_qubits) / 2,
        collective_op(PAULI_Z, num_qubits) / 2,
    )


def collective_op(op: np.ndarray, num_qubits: int) -> np.ndarray:
    """Compute the collective version of a single-qubit qubit operator: sum_q op_q."""
    assert op.shape == (2, 2)
    return sum((act_on_subsystem(num_qubits, op, qubit) for qubit in range(num_qubits)), start=np.array(0))


def act_on_subsystem(num_qubits: int, op: np.ndarray, *qubits: int) -> np.ndarray:
    """Return an operator that acts with 'op' in the given qubits, and trivially (with the identity operator) on all other qubits."""
    assert op.shape == (2 ** len(qubits),) * 2, "Operator shape {op.shape} is inconsistent with the number of target qubits provided, {num_qubits}!"
    identity = np.eye(2 ** (num_qubits - len(qubits)), dtype=op.dtype)
    system_op = np.kron(op, identity)

    # rearrange operator into tensor factors addressing each qubit
    system_op = np.moveaxis(
        system_op.reshape((2,) * 2 * num_qubits),
        range(num_qubits),
        range(0, 2 * num_qubits, 2),
    ).reshape((4,) * num_qubits)

    # move the first len(qubits) tensor factors to the target qubits
    system_op = np.moveaxis(
        system_op,
        range(len(qubits)),
        qubits,
    )

    # split and re-combine tensor factors again to recover the operator as a matrix
    return np.moveaxis(
        system_op.reshape((2,) * 2 * num_qubits),
        range(0, 2 * num_qubits, 2),
        range(num_qubits),
    ).reshape((2**num_qubits,) * 2)


def get_jacobian_func(
    simulate_func: Callable,
    *,
    disable_diffrax: bool = DISABLE_DIFFRAX,
    step_sizes: float | Sequence[float] = 1e-10,
) -> Callable:
    """Convert a simulation method into a function that returns its Jacobian."""

    if not DISABLE_DIFFRAX:

        def get_jacobian(params: Sequence[float], *args: object, **kwargs: object) -> np.ndarray:
            primals, vjp_func = jax.vjp(simulate_func, params, *args)
            result = np.zeros((primals.shape[1], primals.shape[0], len(params)), dtype=COMPLEX_TYPE)
            for ii, jj in numpy.ndindex(primals.shape):
                seed = np.zeros(primals.shape, dtype=COMPLEX_TYPE)
                seed = seed.at[ii, jj].set(1.0)
                res = np.array(vjp_func(seed)[0]).flatten()
                seed = seed.at[ii, jj].set(1.0j)
                res = res + np.array(vjp_func(seed)[0]).flatten() * 1.0j
                # Take the conjugate to account for Jax convention. See discussion:
                # https://github.com/google/jax/issues/4891
                result = result.at[ii, jj, :].set(np.conj(res))
            return result

        return get_jacobian

    def get_jacobian_manually(params: Sequence[float], *args: object, **kwargs: object) -> np.ndarray:
        if isinstance(step_sizes, float):
            param_step_sizes = [step_sizes] * len(params)
        assert len(param_step_sizes) == len(params)

        result_at_params = simulate_func(params, *args, **kwargs)
        shifted_results = []
        for idx, step_size in enumerate(param_step_sizes):
            new_params = list(params)
            new_params[idx] += step_size
            result_at_params_with_step = simulate_func(new_params, *args, **kwargs)
            shifted_results.append((result_at_params_with_step - result_at_params) / step_size)

        return np.stack(shifted_results, axis=-1)

    return get_jacobian_manually


def print_jacobian(jacobian: np.ndarray, precision: int = 3, linewidth: int = 200) -> None:
    np.set_printoptions(precision=precision, suppress=True, linewidth=linewidth)
    params = jacobian.shape[2]
    for pp in range(params):
        print(f"d(final_state/d(params[{pp}]):")
        print(jacobian[:, :, pp])


if __name__ == "__main__":
    # parse arguments
    parser = argparse.ArgumentParser(
        description="Simulate a simple one-axis twisting (OAT) protocol.",
        formatter_class=argparse.MetavarTypeHelpFormatter,
    )
    parser.add_argument("--num_qubits", type=int, default=4)
    parser.add_argument("--dissipation", type=float, default=0.0)
    parser.add_argument("--params", type=float, nargs="+", default=np.array([0.5, 0.5, 0.5, 0]))
    parser.add_argument("--jacobian", action="store_true", default=False)
    args = parser.parse_args(sys.argv[1:])

    if args.jacobian:
        get_jacobian = get_jacobian_func(simulate_OAT)
        jacobian = get_jacobian(args.params, args.num_qubits, dissipation_rates=args.dissipation)
        print_jacobian(jacobian)

    # simulate the OAT protocol
    final_state = simulate_OAT(args.params, args.num_qubits, dissipation_rates=args.dissipation)

    # compute collective Pauli operators
    mean_X = collective_op(PAULI_X, args.num_qubits) / args.num_qubits
    mean_Y = collective_op(PAULI_Y, args.num_qubits) / args.num_qubits
    mean_Z = collective_op(PAULI_Z, args.num_qubits) / args.num_qubits
    mean_ops = [mean_X, mean_Y, mean_Z]

    # print out expectation values and variances
    final_pauli_vals = np.array([(final_state @ op).trace().real for op in mean_ops])
    final_pauli_vars = np.array([(final_state @ (op @ op)).trace().real - mean_op_val**2 for op, mean_op_val in zip(mean_ops, final_pauli_vals)])
    print("[<X>, <Y>, <Z>]:", final_pauli_vals)
    print("[var(X), var(Y), var(Z)]:", final_pauli_vars)



#!/usr/bin/env python3
import matplotlib.pyplot as plt
import numpy as np

from qfi_opt import spin_models

N = 4
dissipation = 1
G = spin_models.collective_op(spin_models.PAULI_Z, N) / (2 * N)

np.random.seed(6)
x0 = np.random.uniform(0, 1, 4)
x1 = np.random.uniform(0, 1, 4)

disp = 1 / 50
vals = np.arange(0, 1 + disp, disp)

results = {}
for i, alpha in enumerate(vals):
    params = alpha * x0 + (1 - alpha) * x1

    rho = spin_models.simulate_OAT(params, N, dissipation_rates=dissipation)

    u, v = np.linalg.eig(rho)
    results[i] = [u, v]

Y = np.array([np.sort(np.real(results[i][0])) for i in range(len(results))])

plt.plot(Y, linewidth=6, alpha=0.8, solid_joinstyle="miter")
plt.xticks([0, len(vals)], ["$x_0$", "$x_1$"])
plt.xlabel("Parameters $x$")
plt.ylabel("Eigenvalues of $\\rho(x)$")
# plt.savefig("Initial.png", dpi=300, bbox_inches="tight", transparent=True)
plt.savefig("kink.png", dpi=300, bbox_inches="tight")
plt.close()



import dataclasses
import functools
from collections.abc import Callable, Sequence

import jax.numpy as np
import numpy

from qfi_opt import spin_models


@dataclasses.dataclass(kw_only=True)
class Transformation:
    """An object representing a sequence of transformations to apply to a quantum state."""

    final_rz: float = 0
    flip_xy: float | None = None
    conjugate: bool = False

    def transform(self, state: np.ndarray) -> np.ndarray:
        num_qubits = spin_models.log2_int(state.shape[0])
        new_state = state.copy()
        if self.final_rz:
            # apply a global spin rotation 'Rz(self.final_rz)'
            phase_mat = rot_z_mat(num_qubits, self.final_rz)
            new_state = phase_mat * new_state
        if self.flip_xy is not None:
            # apply a global spin rotation by an angle 'pi' about a specified axis in the X-Y plane
            phase_mat = rot_z_mat(num_qubits, 2 * np.pi * self.flip_xy)
            new_state = phase_mat * (phase_mat.conj() * new_state)[::-1, ::-1]
        if self.conjugate:
            # complex conjugate the state
            new_state = new_state.conj()
        return new_state


# Type signature for the parameters of a sensing protocol.
Params = tuple[float, float, float, float, float]

# A symmetry `S` maps a set of parameters `p_old` to a set of parameters `p_new` together with a transformation `T`, i.e., `S: p_old --> (p_new, T)`.
# The symmetry is such that the following two procedures yield the same state:
# - Running a sensing protocol with parameters `p_old`.
# - Running the same sensing protocol with parameters `p_new`, and applying the transformation `T` to the resulting state.
# In other words, if `rho(p)` is the state returned by a sensing protocol with parameters `p`, then `rho(p_old) = T(rho(p_new))`.
Symmetry = Callable[[float, float, float, float, float], tuple[Params, Transformation]]


def get_random_hamiltonian(dim: int) -> np.ndarray:
    """Construct a random Hamiltonian on a system of with the given dimension."""
    ham = numpy.random.random((dim, dim)) + 1j * numpy.random.random((dim, dim))
    return np.array(ham + ham.conj().T) / 2


@functools.cache
def rot_z_mat(num_qubits: int, angle: float) -> np.ndarray:
    """Construct the matrix 'phase_mat' for which 'phase_mat * density_op' rotates the state 'density_op' about the Z axis by the given angle."""
    if not angle:
        return np.ones((2**num_qubits,) * 2)
    _, _, collective_Sz = spin_models.collective_spin_ops(num_qubits)
    phase_vec = np.exp(-1j * angle * collective_Sz.diagonal())
    return phase_vec[:, np.newaxis] * np.conj(phase_vec)


def get_symmetries_common() -> Sequence[Symmetry]:
    """Generate a list of parameter symmetries common to all protocols.

    Assumes zero dissipation.

    Each symmetry is a map from 'old_params' --> '(new_params, transformation)', where 'transformation' indicates how a quantum state prepared with the
    'new_params' should be additionally transformed to recover an exact symmetry.  Note that the additional transformations (should) have no effect on the QFI.
    """

    def reflect_1(t_1: float, a_1: float, t_ent: float, t_2: float, a_2: float) -> tuple[Params, Transformation]:
        return (-t_1, a_1 + 0.5, t_ent, t_2, a_2), Transformation()

    def reflect_2(t_1: float, a_1: float, t_ent: float, t_2: float, a_2: float) -> tuple[Params, Transformation]:
        return (t_1, a_1, t_ent, -t_2, a_2 + 0.5), Transformation()

    def shift_2(t_1: float, a_1: float, t_ent: float, t_2: float, a_2: float) -> tuple[Params, Transformation]:
        return (t_1, a_1, t_ent, t_2 + 1, a_2), Transformation(flip_xy=a_2)

    return [reflect_1, reflect_2, shift_2]


def get_symmetries_U1_Z2() -> Sequence[Symmetry]:
    """Generate a list of symmetries for protocol with an axial U(1) and transverse Z_2 symmetry."""

    def shift_axes(t_1: float, a_1: float, t_ent: float, t_2: float, a_2: float) -> tuple[Params, Transformation]:
        axis_shift = numpy.random.random()
        final_rz = 2 * np.pi * axis_shift
        return (t_1, a_1 - axis_shift, t_ent, t_2, a_2 - axis_shift), Transformation(final_rz=final_rz)

    def shift_1(t_1: float, a_1: float, t_ent: float, t_2: float, a_2: float) -> tuple[Params, Transformation]:
        return (t_1 + 1, a_1, t_ent, t_2, 2 * a_1 - a_2), Transformation(flip_xy=a_1)

    def conjugate(t_1: float, a_1: float, t_ent: float, t_2: float, a_2: float) -> tuple[Params, Transformation]:
        return (-t_1, -a_1, -t_ent, -t_2, -a_2), Transformation(conjugate=True)

    return [shift_axes, shift_1, conjugate]


def get_symmetries_Z2_Z2() -> Sequence[Symmetry]:
    """Generate a list of symmetries for protocol with an axial Z_2 and transverse Z_2 symmetry."""

    def rot_z(t_1: float, a_1: float, t_ent: float, t_2: float, a_2: float) -> tuple[Params, Transformation]:
        final_rz = -np.pi / 2
        return (t_1, a_1 + 0.25, -t_ent, t_2, a_2 + 0.25), Transformation(final_rz=final_rz)

    def pi_x(t_1: float, a_1: float, t_ent: float, t_2: float, a_2: float) -> tuple[Params, Transformation]:
        final_rz = 4 * np.pi * a_2
        return (t_1 + 1, -a_1, -t_ent, t_2 + 1, -a_2), Transformation(final_rz=final_rz)

    def conjugate(t_1: float, a_1: float, t_ent: float, t_2: float, a_2: float) -> tuple[Params, Transformation]:
        return (-t_1, -a_1, t_ent, -t_2, -a_2), Transformation(conjugate=True)

    return [rot_z, pi_x, conjugate]


def get_symmetries_OAT(even_qubit_number: bool) -> Sequence[Symmetry]:
    """Generate a list of symmetries unique to the OAT protocol."""

    def shift_ent(t_1: float, a_1: float, t_ent: float, t_2: float, a_2: float) -> tuple[Params, Transformation]:
        final_rz = np.pi * even_qubit_number
        return (t_1, a_1, t_ent + 1, t_2, a_2 + 0.5 * even_qubit_number), Transformation(final_rz=final_rz)

    return [shift_ent]


def run_symmetry_tests(simulate_method: Callable[[Sequence[float]], np.ndarray], symmetries: Sequence[Symmetry], atol: float = 1e-6) -> None:
    """Test that the given simulation method obeys the given symmetries."""
    params = tuple(numpy.random.random(5))
    state = simulate_method(params)
    for symmetry in symmetries:
        new_params, transformation = symmetry(*params)
        new_state = simulate_method(new_params)
        assert np.allclose(state, transformation.transform(new_state), atol=atol)


def test_symmetries() -> None:
    """Test the symmetry transformations that we use to cut down the parameter domain of sensing protocols."""
    for _ in range(5):  # test several random instances
        coupling_op = get_random_hamiltonian(4)
        coupling_exponent = numpy.random.random() * 3

        for num_qubits in [2, 3]:  # test both even and odd qubit numbers
            # test common symmetries
            run_symmetry_tests(
                lambda params: spin_models.simulate_spin_chain(params, num_qubits, coupling_op, coupling_exponent),
                get_symmetries_common(),
            )

            # test U(1) x Z_2 symmetries
            for simulate_method in [
                lambda params: spin_models.simulate_ising_chain(params, num_qubits, coupling_exponent),
                lambda params: spin_models.simulate_XX_chain(params, num_qubits, coupling_exponent),
            ]:
                run_symmetry_tests(simulate_method, get_symmetries_U1_Z2())

            # test Z_2 x Z_2 symmetries
            run_symmetry_tests(
                lambda params: spin_models.simulate_local_TAT_chain(params, num_qubits, coupling_exponent),
                get_symmetries_Z2_Z2(),
            )

            # test OAT symmetries
            run_symmetry_tests(
                lambda params: spin_models.simulate_OAT(params, num_qubits),
                get_symmetries_OAT(num_qubits % 2 == 0),
            )


def run_derivatives_tests(simulate_method: Callable[[Sequence[float]], np.ndarray], symmetries: Sequence[Symmetry], atol: float = 1e-6) -> None:
    """Test that the given simulation method obeys the given symmetries."""
    params = tuple(numpy.random.random(5))
    state = simulate_method(params)
    for symmetry in symmetries:
        new_params, transformation = symmetry(*params)
        new_state = simulate_method(new_params)
        assert np.allclose(state, transformation.transform(new_state), atol=atol)


def test_derivatives() -> None:
    """Test the symmetry transformations that we use to cut down the parameter domain of sensing protocols."""
    for _ in range(5):  # test several random instances
        coupling_op = get_random_hamiltonian(4)
        coupling_exponent = numpy.random.random() * 3

        for num_qubits in [2, 3]:  # test both even and odd qubit numbers
            # test derivatives of common symmetries
            run_derivatives_tests(
                lambda params: spin_models.simulate_spin_chain(params, num_qubits, coupling_op, coupling_exponent),
                get_symmetries_common(),
            )

            # test derivatives of U(1) x Z_2 symmetries
            for simulate_method in [
                lambda params: spin_models.simulate_ising_chain(params, num_qubits, coupling_exponent),
                lambda params: spin_models.simulate_XX_chain(params, num_qubits, coupling_exponent),
            ]:
                run_derivatives_tests(simulate_method, get_symmetries_U1_Z2())

            # test derivatives of Z_2 x Z_2 symmetries
            run_derivatives_tests(
                lambda params: spin_models.simulate_local_TAT_chain(params, num_qubits, coupling_exponent),
                get_symmetries_Z2_Z2(),
            )

            # test derivatives of OAT symmetries
            run_derivatives_tests(
                lambda params: spin_models.simulate_OAT(params, num_qubits),
                get_symmetries_OAT(num_qubits % 2 == 0),
            )



#!/usr/bin/env python3
import os
import pickle
import sys

import nlopt
import numpy as np
from ibcdfo.pounders import pounders
from ibcdfo.pounders.general_h_funs import identity_combine as combinemodels
from mpi4py import MPI

import qfi_opt
from qfi_opt import spin_models
from qfi_opt.examples.calculate_qfi import compute_eigendecompotion, compute_QFI

root_dir = os.path.dirname(os.path.dirname(qfi_opt.__file__))
minq5_dir = os.path.join(root_dir, "minq", "py", "minq5")
if not os.path.isdir(minq5_dir):
    messages = [
        "Please install (or symlink) MINQ in the QFI-Opt project directory.",
        "You can do this with:",
        f"\n  git clone https://github.com/POptUS/MINQ.git {root_dir}/minq\n",
        "Or, if you already have MINQ somewhere, run:",
        f"\n  ln -s <minq-path> {root_dir}/minq\n",
    ]
    exit("\n".join(messages))
sys.path.append(minq5_dir)

orbit_dir = os.path.isdir(os.path.join(root_dir, "orbit", "py"))
orbit_found = os.path.isdir(orbit_dir)
if not orbit_found:
    messages = [
        "WARNING: 'orbit' not found",
        "If you already have orbit somewhere, run:",
        f"\n  ln -s <orbit-path> {root_dir}/orbit\n",
    ]
    print("\n".join(messages))
else:
    sys.path.append(orbit_dir)
    from orbit4py import ORBIT2


def sim_wrapper(x, grad, obj, obj_params):
    """Wrapper for `nlopt` that creates and updates a database of simulation inputs/outputs.

    Note that for large databases (or fast simulations), the database lookup can be more expensive than performing the simulation.
    """
    use_DB = False
    match = 0
    if use_DB:
        # Look through the database to see if there is a match
        database = obj.__name__ + "_" + str(obj_params["N"]) + "_" + str(obj_params["dissipation"]) + "_database.npy"
        DB = []
        if os.path.exists(database):
            DB = np.load(database, allow_pickle=True)
            for db_entry in DB:
                if np.allclose(db_entry["var_vals"], x, rtol=1e-12, atol=1e-12):
                    rho = db_entry["rho"]
                    match = 1
                    break

    if match == 0:
        # Do the sim
        rho = obj(x, obj_params["N"], dissipation_rates=obj_params["dissipation"])

        if use_DB:
            # Update database
            to_save = {"rho": rho, "var_vals": x}
            DB = np.append(DB, to_save)
            np.save(database, DB)

    vals, vecs = compute_eigendecompotion(rho)
    qfi = compute_QFI(vals, vecs, obj_params["G"])
    print(x, qfi, flush=True)
    all_f.append(qfi)
    all_X.append(x)
    return -1 * qfi  # negative because we are maximizing


def run_orbit(obj, obj_params, n, x0):
    calfun = lambda x: sim_wrapper(x, [], obj, obj_params)
    gtol = 1e-9  # Gradient tolerance used to stop the local minimization [1e-5]
    rbftype = "cubic"  # Type of RBF (multiquadric, cubic, Gaussian) ['cubic']
    npmax = 2 * n + 1  # Maximum number of interpolation points [2*n+1]
    trnorm = 0  # Type f trust-region norm [0]
    Low = -5000 * np.ones(n)  # 1-by-n Vector of lower bounds [zeros(1,n)]
    Upp = 5000 * np.ones(n)  # 1-by-n Vector of upper bounds [ones(1,n)]
    gamma_m = 0.5  # Reduction factor = factor of the LHS points you'd start a local run from [.5]
    maxdelta = np.inf
    delta = 1
    nfs = 1

    xkin = 0
    X = np.array(x0)
    F = np.array(calfun(X))

    X, F, xkin, *_ = ORBIT2(calfun, rbftype, gamma_m, n, max_evals, npmax, delta, maxdelta, trnorm, gtol, Low, Upp, nfs, X, F, xkin)

    # print("optimum at ", X[xkin])
    # print("minimum value = ", F[xkin])

    return F[xkin], X[xkin]


def run_pounder(obj, obj_params, n, x0):
    calfun = lambda x: sim_wrapper(x, [], obj, obj_params)
    X = np.array(x0)
    F = np.array(calfun(X))
    Low = -np.inf * np.ones((1, n))
    Upp = np.inf * np.ones((1, n))
    mpmax = 2 * n + 1
    delta = 0.1
    m = 1
    nfs = 1
    printf = False
    spsolver = 2
    gtol = 1e-9
    xind = 0
    hfun = lambda F: F

    X, F, _, xkin = pounders(calfun, X, n, mpmax, max_evals, gtol, delta, nfs, m, F, xind, Low, Upp, printf, spsolver, hfun, combinemodels)

    # print("optimum at ", X[xkin])
    # print("minimum value = ", F[xkin])

    return F[xkin], X[xkin]


def run_nlopt(obj, obj_params, num_params, x0, solver):
    opt = nlopt.opt(getattr(nlopt, solver), num_params)
    # opt = nlopt.opt(nlopt.LN_NELDERMEAD, num_params)  # Doesn't use derivatives and will work
    # opt = nlopt.opt(nlopt.LD_MMA, num_params) # Needs derivatives to work. Without grad being set (in-place) it is zero, so first iterate is deemed stationary

    opt.set_min_objective(lambda x, grad: sim_wrapper(x, grad, obj, obj_params))
    opt.set_xtol_rel(1e-4)
    opt.set_maxeval(max_evals)

    # # Because the objective is periodic, don't set bounds (but don't need to sample so much)
    # opt.set_lower_bounds(lb)
    # opt.set_upper_bounds(ub)

    x = opt.optimize(x0)
    minf = opt.last_optimum_value()
    # print("optimum at ", x)
    # print("minimum value = ", minf)
    # print("result code = ", opt.last_optimize_result())

    return minf, x


if __name__ == "__main__":  # noqa: C901 # ignore "complexity" check for the code below
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    N = 4
    G = spin_models.collective_op(spin_models.PAULI_Z, N) / (2 * N)

    for dissipation_rate in np.append([0], np.linspace(0.1, 5, 20)):
        obj_params = {}
        obj_params["N"] = N
        obj_params["dissipation"] = dissipation_rate
        obj_params["G"] = G

        max_evals = 10

        for num_params in [4, 5]:
            lb = np.zeros(num_params)
            ub = np.ones(num_params)

            for seed in range(size):
                if seed % size == rank:
                    # x0 = 0.5 * np.ones(num_params)  # This is an optimum for the num_params==4 problems
                    np.random.seed(seed)
                    x0 = np.random.uniform(lb, ub, num_params)

                    match num_params:
                        case 4:
                            models = ["simulate_OAT", "simulate_ising_chain", "simulate_XX_chain"]
                            # models = ["simulate_OAT"]
                        case 5:
                            models = ["simulate_TAT", "simulate_local_TAT_chain"]
                            # models = ["simulate_TAT"]

                    for model in models:
                        filename = model + "_" + str(dissipation_rate) + "_" + str(seed) + ".pkl"
                        fig_filename = "Results_" + model + "_" + str(dissipation_rate) + "_" + str(seed)
                        if os.path.exists(filename):
                            continue
                        if os.path.exists(fig_filename + ".png"):
                            continue
                        obj = getattr(spin_models, model)

                        best_val = {}
                        best_pt = {}
                        for solver in ["LN_NELDERMEAD", "LN_BOBYQA", "ORBIT", "POUNDER"]:
                            global all_f, all_X
                            all_f = []
                            all_X = []
                            if solver in ["LN_NELDERMEAD", "LN_BOBYQA"]:
                                run_nlopt(obj, obj_params, num_params, x0, solver)
                            elif solver in ["ORBIT"]:
                                if not orbit_found:
                                    continue
                                run_orbit(obj, obj_params, num_params, x0)
                            elif solver in ["POUNDER"]:
                                run_pounder(obj, obj_params, num_params, x0)

                            print(all_f, all_X, flush=True)

                            ind = np.argmax(all_f)
                            best_val[solver] = all_f[ind]
                            best_pt[solver] = all_X[ind]

                        best_method = max(best_val, key=best_val.get)
                        # print(model, d, seed, best_val[best_method], best_pt[best_method], flush=True)
                        dic = {}
                        dic["best_method"] = best_method
                        dic["best_val"] = best_val[best_method]
                        dic["best_pt"] = best_pt[best_method]

                        with open(filename, "wb") as f:
                            pickle.dump(dic, f)

                        sys.exit("a")

                        #     plt.figure(fig_filename)
                        #     plt.plot(all_f, label=solver)

                        #     for i in range(1, len(all_f)):
                        #         all_f[i] = max(all_f[i - 1], all_f[i])

                        #     plt.figure(fig_filename + "best")
                        #     plt.plot(all_f, label=solver)

                        # plt.figure(fig_filename)
                        # plt.legend()
                        # plt.title(fig_filename)
                        # plt.savefig(fig_filename + ".png", dpi=300)

                        # plt.figure(fig_filename + "best")
                        # plt.legend()
                        # plt.title(fig_filename)
                        # plt.savefig(fig_filename + "best" + ".png", dpi=300)



#!/usr/bin/env python3
import matplotlib.pyplot as plt

import qfi_opt

params = [0.85255953, 1.0, 2.75901551, 1.66809236, 1.00000226]
num_qubits = 4
dissipation_rates = 0

state = qfi_opt.spin_models.simulate_TAT(params, num_qubits, dissipation_rates=dissipation_rates)

cat_state_fidelity = 0.5 * sum(abs(state[ii, jj]) for ii in [0, -1] for jj in [0, -1])
print("cat state fidelity:", cat_state_fidelity)


qfi_opt.plot.husimi(state)
qfi_opt.plot.histogram(state)
plt.show()



#!/usr/bin/env python3
import numpy as np

from qfi_opt import spin_models


def variance(rho: np.ndarray, G: np.ndarray) -> float:
    """Variance of self-adjoint operator (observable) G in the state rho."""
    return (G @ G @ rho).trace().real - (G @ rho).trace().real ** 2


def compute_eigendecomposition(rho: np.ndarray):
    # Compute eigendecomposition for rho
    eigvals, eigvecs = np.linalg.eigh(rho)
    eigvecs = eigvecs.T  # make the k-th eigenvector eigvecs[k, :] = eigvecs[k]
    return eigvals, eigvecs


def compute_QFI(
    eigvals: np.ndarray, eigvecs: np.ndarray, G: np.ndarray, A: np.ndarray, dA: np.ndarray, d2A: np.ndarray, grad, tol: float = 1e-8, etol_scale: float = 10
) -> float:
    # Note: The eigenvectors must be rows of eigvecs
    num_vals = len(eigvals)
    num_params = dA.shape[0]

    # There should never be negative eigenvalues, so their magnitude gives an
    # empirical estimate of the numerical accuracy of the eigendecomposition.
    # We discard any QFI terms denominators within an order of magnitude of
    # this value.
    tol = max(tol, -etol_scale * np.min(eigvals))

    # Compute QFI and grad
    running_sum = 0

    if grad.size > 0:
        grad[:] = np.zeros(num_params)
        # compute gradients of each eigenvalue
        # lambda_grads, psi_grads, eigvecs = get_matrix_grads(A, dA, d2A, eigvals, eigvecs, tol)
        lambda_grads, psi_grads = get_matrix_grads_lazy(A, dA, eigvals, eigvecs)

    for i in range(num_vals):
        for j in range(i + 1, num_vals):
            denom = eigvals[i] + eigvals[j]
            if not np.isclose(denom, 0, atol=tol, rtol=tol):
                numer = (eigvals[i] - eigvals[j]) ** 2
                term = eigvecs[i].conj() @ G @ eigvecs[j]
                quotient = numer / denom
                squared_modulus = np.absolute(term) ** 2
                running_sum += quotient * squared_modulus
                if grad.size > 0:
                    for k in range(num_params):
                        # fill in gradient
                        grad[k] += kth_partial_derivative(
                            quotient,
                            squared_modulus,
                            eigvals[i],
                            eigvals[j],
                            lambda_grads[k, i],
                            lambda_grads[k, j],
                            eigvecs[i],
                            eigvecs[j],
                            psi_grads[k, i],
                            psi_grads[k, j],
                            G,
                        )

    if grad.size > 0:
        return 4 * running_sum, 4 * grad
    else:
        return 4 * running_sum, []


def get_matrix_grads_lazy(A, dA, eigvals, eigvecs):
    num_vals = len(eigvals)
    num_params = dA.shape[0]
    lambda_grads = np.zeros((num_params, num_vals))
    psi_grads = np.zeros((num_params, num_vals, num_vals), dtype="cdouble")

    for k in range(num_params):
        for index in range(num_vals):
            dlambda, dpsi = matrixder_lazy(A, dA[k], eigvals[index], eigvecs[index])

            lambda_grads[k, index] = np.real(dlambda)
            psi_grads[k, index] = dpsi.flatten()

    return lambda_grads, psi_grads


def get_matrix_grads(A, dA, d2A, eigvals, eigvecs, tol):
    num_vals = len(eigvals)
    num_params = dA.shape[0]
    lambda_grads = np.zeros((num_params, num_vals))
    psi_grads = np.zeros((num_params, num_vals, num_vals), dtype="cdouble")
    updated_eigvecs = np.zeros((num_vals, num_vals), dtype="cdouble")

    for k in range(num_params):
        already_used = []
        for val in range(num_vals):
            if val not in already_used:
                # check for multiplicity within tolerance
                eigval = eigvals[val]
                indices = np.where(np.abs(eigvals.flatten() - eigval) <= tol)[0]
                already_used = already_used + indices.tolist()

                dlambda_k, dV_k, V_k = matrixder(A, dA[k], d2A[k], eigval, eigvecs[indices])

                lambda_grads[k, indices] = dlambda_k
                psi_grads[k, indices] = dV_k
                updated_eigvecs[indices] = V_k

    return lambda_grads, psi_grads, updated_eigvecs


def matrixder_lazy(A, dA, eigval, eigvec):
    # build the big matrix
    num_vals = A.shape[0]
    Eye = np.eye(num_vals)
    eigvec = eigvec[:, np.newaxis]
    W = np.concatenate((A - eigval * Eye, -1.0 * eigvec), axis=1, dtype="cdouble")
    row = np.append(eigvec.conj().T, 0)
    row = row[:, np.newaxis].T
    W = np.concatenate((W, row), axis=0, dtype="cdouble")

    # right hand side
    rhs = -1.0 * dA @ eigvec
    rhs = np.append(rhs, 0)
    rhs = rhs[:, np.newaxis]
    sol = np.linalg.solve(W, rhs)

    # read off values
    dlambda = sol[num_vals]
    dpsi = sol[:num_vals]

    return dlambda, dpsi


def matrixder(A, dA, ddA, lam, arbV):
    # A is a Hermitian n times n matrix
    # dA is the first derivative of A with respect to a parameter of interest
    # ddA is the second derivative of A with respect to a parameter of interest
    # lam is one eigenvalue of A
    # arbV is an orthonormal basis for the eigenspace associated with lam, of shape n x r (r is multiplicity)

    # because arbV is transposed
    arbV = arbV.T

    # Step 1: Initialize a few structures
    n, r = arbV.shape
    B = np.eye(n)

    # Step 2: set up initial eigenvalue problem
    M = arbV.conj().T @ dA @ arbV
    W = A - lam * B

    # Step 3: (we only want first derivatives, so Step 3 is not necessary)

    # Step 4: Solve the eigenvalue problem
    dLambda, U = np.linalg.eigh(M)
    V = arbV @ U  # this is V in the "correct" basis

    # Step 5: reduced eigenspace
    V1 = np.linalg.solve(W, -dA @ arbV + V @ np.diag(dLambda) @ U.conj().T)

    # Step 6: Compute M2
    if r > 1:  # only do this step if repeated eigenvalue
        M2 = V.conj().T @ ddA @ V + 2.0 * V.conj().T @ (-V1 @ U @ np.diag(dLambda) + dA @ V1 @ U)

    # Step 7: Build C
    D = -1.0 * V.conj().T @ V1 @ U
    C = np.zeros((r, r), dtype="cdouble")
    for i in range(r):
        for j in range(r):
            if i == j:
                C[i, j] = D[i, j]
            else:
                C[i, j] = M2[i, j] / (2 * (dLambda[j] - dLambda[i]))

    # Step 8: Derivatives of eigenvectors
    dV = V1 @ U + V @ C

    return dLambda, dV.T, V.T


def quotient_partial_derivative(lambda_i, lambda_j, d_lambda_i, d_lambda_j):
    squared_diff = (lambda_i - lambda_j) ** 2
    fprimeg = 2 * squared_diff * (d_lambda_i - d_lambda_j)
    gprimef = squared_diff * (d_lambda_i + d_lambda_j)
    der = (fprimeg - gprimef) / (lambda_i + lambda_j) ** 2

    return der


def modulus_partial_derivative(psi_i, psi_j, d_psi_i, d_psi_j, G):
    inner_product = psi_i.conj() @ G @ psi_j
    left_derivative = d_psi_i.conj() @ G @ psi_j
    right_derivative = psi_i.conj() @ G @ d_psi_j

    real_der = 2 * inner_product.real * (left_derivative.real + right_derivative.real)
    imag_der = 2 * inner_product.imag * (left_derivative.imag + right_derivative.imag)

    der = real_der + imag_der

    return der


def kth_partial_derivative(quotient, modulus, lambda_i, lambda_j, d_lambda_i, d_lambda_j, psi_i, psi_j, d_psi_i, d_psi_j, G):
    quotient_der = quotient_partial_derivative(lambda_i, lambda_j, d_lambda_i, d_lambda_j)
    modulus_der = modulus_partial_derivative(psi_i, psi_j, d_psi_i, d_psi_j, G)

    der = quotient * modulus_der + modulus * quotient_der

    return der


if __name__ == "__main__":
    num_spins = 4
    dissipation = 0
    op = spin_models.collective_op(spin_models.PAULI_Z, num_spins) / (2 * num_spins)

    num_rand_pts = 2
    print_precision = 6
    # Calculate QFI for models at random points in the domain.
    for num_params in [4, 5]:
        match num_params:
            case 4:
                models = ["simulate_OAT", "simulate_ising_chain", "simulate_XX_chain"]
            case 5:
                models = ["simulate_TAT", "simulate_local_TAT_chain"]

        for model in models:
            print(model)
            np.random.seed(0)
            obj = getattr(spin_models, model)
            get_jacobian = spin_models.get_jacobian_func(obj)

            params = 0.5 * np.ones(num_params)
            rho = obj(params, num_spins, dissipation_rates=dissipation)

            grad_of_rho = get_jacobian(params, num_spins, dissipation_rates=dissipation)
            spin_models.print_jacobian(grad_of_rho, precision=print_precision)
            vals, vecs = compute_eigendecomposition(rho)

            qfi = compute_QFI(vals, vecs, op)
            print(f"QFI is {qfi} for {params}")

            params[-1] = 0.0
            rho = obj(params, num_spins, dissipation_rates=dissipation)

            grad_of_rho = get_jacobian(params, num_spins, dissipation_rates=dissipation)
            spin_models.print_jacobian(grad_of_rho, precision=print_precision)
            vals, vecs = compute_eigendecomposition(rho)

            qfi = compute_QFI(vals, vecs, op)
            print(f"QFI is {qfi} for {params}")

            params[-1] = 1.0
            rho = obj(params, num_spins, dissipation_rates=dissipation)

            grad_of_rho = get_jacobian(params, num_spins, dissipation_rates=dissipation)
            spin_models.print_jacobian(grad_of_rho, precision=print_precision)
            vals, vecs = compute_eigendecomposition(rho)

            qfi = compute_QFI(vals, vecs, op)
            print(f"QFI is {qfi} for {params}")
